(module disjoint-set
  (
   disjoint-set-find!
   disjoint-set-node-parenti
   disjoint-set-node?
   disjoint-set-ref
   disjoint-set-root-rank
   disjoint-set-root?
   disjoint-set-set!
   disjoint-set-size
   disjoint-set-size-set!
   disjoint-set-unite!
   disjoint-set?
   make-disjoint-set
   make-disjoint-set-node
   make-disjoint-set-root
   )

  (import
    scheme
    (srfi 8)

    (only (chicken base)
          fixnum?
          when)
    )

  (define %disjoint-set% 'disjoint-set)

  (define (make-disjoint-set-node parenti) parenti)
  (define (make-disjoint-set-root rank) `(,rank)) ; A root's parent is itself

  (define disjoint-set-node? fixnum?)
  (define disjoint-set-root? pair?)

  (define (disjoint-set-node-parenti parenti) parenti)
  (define disjoint-set-root-rank car)

  (define disjoint-set-container-ref vector-ref)
  (define disjoint-set-container-set! vector-set!)

  (define (make-disjoint-set-container% nelems)
    (let ((ret (make-vector nelems)))
      (let loop ((i 0))
        (when (< i nelems)
          (disjoint-set-container-set! ret i (make-disjoint-set-root 0))
          (loop (+ i 1))))
      ret))

  (define (make-disjoint-set% size container) `#(,%disjoint-set% ,size ,container))
  (define (disjoint-set-tag% ds) (vector-ref ds 0))
  (define (disjoint-set-size ds) (vector-ref ds 1))
  (define (disjoint-set-container% ds) (vector-ref ds 2))

  (define (disjoint-set-size-set! ds size)
    (vector-set! ds 1 size)
    ds)

  (define (make-disjoint-set nelems)
    (make-disjoint-set% nelems (make-disjoint-set-container% nelems)))

  (define (disjoint-set? obj)
    (and (vector? obj) (eq? %disjoint-set% (disjoint-set-tag% obj))))
  (define (disjoint-set-ref ds xi)
    (disjoint-set-container-ref (disjoint-set-container% ds) xi))
  (define (disjoint-set-set! ds xi x)
    (disjoint-set-container-set! (disjoint-set-container% ds) xi x)
    ds)

  (define (disjoint-set-find! ds xi)
    (let ((x (disjoint-set-ref ds xi)))
      (if (disjoint-set-root? x)
        (values xi x ds)
        (receive (ri r ds) (disjoint-set-find! ds (disjoint-set-node-parenti x))
          (let ((ds (disjoint-set-set! ds xi (make-disjoint-set-node ri))))
            (values ri r ds))))))

  (define (choose-root&child xi x yi y)
    (let ((x-rank (disjoint-set-root-rank x))
          (y-rank (disjoint-set-root-rank y)))
      (if (< x-rank y-rank)
        (values yi y y-rank xi x x-rank)
        (values xi x x-rank yi y y-rank))))

  (define (disjoint-set-unite! ds xi yi)
    (if (= xi yi)
      ds
      (receive (xi x ds) (disjoint-set-find! ds xi)
        (receive (yi y ds) (disjoint-set-find! ds yi)
          (if (= xi yi) ; both elements belong to the same subset already
            ds
            (receive (xi x x-rank yi y y-rank) (choose-root&child xi x yi y)
              (let ((x-rank (if (= x-rank y-rank) (+ x-rank 1) x-rank)))
                (let* ((ds (disjoint-set-set! ds xi (make-disjoint-set-root x-rank)))
                       (ds (disjoint-set-set! ds yi (make-disjoint-set-node xi)))
                       ; We joined 2 subsets so we have one less subset than before
                       (ds (disjoint-set-size-set! ds (- (disjoint-set-size ds) 1))))
                  ds))))))))
  )
